#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2021
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
#------------------------------------------------------------------------


###############################################################################
#
# Run mambo simulator for the specified number of minutes
#
# This works by running a chunk of instructions and then sampling how much
# time has past.  Once the requested number of minutes have past, the
# procedure will return.  Due to the variable rate of instruction
# execution in mambo, the procedure adjusts how many instructions to run for
# a chunk based on how long it took to run the last chunk of instructions.
#
# Returns 0 for success
#
###############################################################################
proc run_minutes { minutes sim } {
    set cur_time [clock seconds]
    set end_time [expr ($minutes * 60) + $cur_time]

    # Set our initial inst_per_second guestimate
    set inst_per_second 50000000

    # Target polling period in seconds
    set polling_period 10

    # +/- tolerance in seconds for reaching the target polling period
    set tolerance 2

    # Gain for increasing and decreasing inst_per_second when error is
    # outside of tolerance (number of instructions).
    set ramp_up 2000000
    set ramp_dn 2000000

    while { $cur_time < $end_time } {
        set prev_time $cur_time

        # Run simulation for the polling period
        $sim step [expr $inst_per_second * $polling_period]

        # exit early if we stopped for any reason other than
        # reaching the requested instruction count
        set stop_reason [simstop status]
        if { [string first "instruction count" $stop_reason] != 0 } {

            # failure :(
            puts "Exiting timed run early due to $stop_reason"
            return -1
        }

        set cur_time [clock seconds]
        set seconds_taken [expr $cur_time - $prev_time]

        if { $seconds_taken > [expr $polling_period + $tolerance]} {

            # Reduce inst_per_second if we're over our polling period
            set error_seconds [expr $seconds_taken - $polling_period]
            set inst_per_second [expr $inst_per_second - ($error_seconds * $ramp_dn)]

            # don't allow $inst_per_second drop below 1000000
            if { $inst_per_second < 1000000 } {
                set inst_per_second 1000000
            }
        } elseif { $seconds_taken < [expr $polling_period - $tolerance]} {

            # Increase inst_per_second if we're under our polling period
            set error_seconds [expr $polling_period - $seconds_taken]
            set inst_per_second [expr $inst_per_second + ($error_seconds * $ramp_up)]
        }

        # Nice for debugging
        #puts "Last polling period was $seconds_taken.  New inst_per_second is $inst_per_second"
    }

    # success :)
    return 0
}


###############################################################################
#
# Returns the resident memory usage of the running process in kilobytes
#
###############################################################################
proc current_memory_usage { } {
    set myPid [pid];
    return [exec ps --pid=$myPid -o "rss="]
}


###############################################################################
#
# Check for memory leak by running for the requested number of minutes and
# watching how much the memory footprint grows.
#
# The default behavior if no threshold value is provided is to fail if
# the memory footprint grows by more than twice the starting value.
#
###############################################################################
proc check_for_mem_leak { check_minutes sim {threshold 0} } {

    puts "\n********** Running memory leak test for $check_minutes minutes **********\n"

    # Grab 1st snapshot of memory usage
    set first_ss [current_memory_usage]

    # Run for requested number of minutes
    if { [run_minutes $check_minutes $sim] != 0 } {
        puts "Failed to run for specified time ($check_minutes minutes).  Unable to finish memory leak test."
        return
    }

    # Grab 2nd snapshot of memory usage
    set second_ss [current_memory_usage]

    # if no threshold specified or set to 0, use default setting.
    if { $threshold == 0 } {
        # Allow memory to grow up to twice as large as when we started
        # before flagging an error.
        set threshold [expr 2 * $first_ss]
    }

    # Check that 2nd snapshot is no more than 2x 1st snapshot
    if { $second_ss > $threshold } {
        puts "\n****** Possible Memory Leak Detected!! ******\n"
    } else {
        puts "\n******     No Memory Leak Detected     ******\n"
    }
    puts "---------------------------------------------"
    puts "Run Time:       $check_minutes minutes"
    puts "Initial memory: $first_ss KB"
    puts "Final memory:   $second_ss KB"
    puts "---------------------------------------------\n"
}


